#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This module defines unit tests for `pt.ptutils.cli.simple`.
"""

from pt.ptutils.cli.simple import (
    arg,
    shared_arg,
    use_shared_arg,
    shared_arg_exists,
    remove_shared_arg,
    arg_group,
    use_arg_group,
    arg_group_exists,
    remove_arg_group,
    command,
    command_exists,
    remove_command,
    get_argument,
    has_argument
)


# ------------------------------------------------------------------------------------------------------------------------
def test_cli_simple_create_command():
    """ Test ability to create a command using the `pt.ptutils.cli.simple.command` decorator. """
    try:
        @command
        def my_command():
            pass
        assert command_exists('my_command')
    finally:
        remove_command('my_command')
        assert not command_exists('my_command')


# ------------------------------------------------------------------------------------------------------------------------
def test_cli_simple_create_arg_group():
    """ Test ability to create an argument group using the `pt.ptutils.cli.simple.arg_group` decorator. """
    try:
        @arg_group
        def fruits():
            pass
        assert arg_group_exists('fruits')
    finally:
        remove_arg_group('fruits')
        assert not arg_group_exists('fruits')


# ------------------------------------------------------------------------------------------------------------------------
def test_cli_simple_create_shared_arg():
    """ Test ability to create a shared argument using the `pt.ptutils.cli.simple.shared_arg` function. """
    try:
        shared_arg(
            '-w', '--whatchamacallits',
            dest = "whatchamacallits"
        )
        assert shared_arg_exists('whatchamacallits')
    finally:
        remove_shared_arg('whatchamacallits')
        assert not shared_arg_exists('whatchamacallits')


# ------------------------------------------------------------------------------------------------------------------------
def test_cli_simple_use_shared_arg():
    """ Test ability to use a shared argument using the `pt.ptutils.cli.simple.use_shared_arg` function. """
    try:
        # define shared arg
        batch = shared_arg(
            '-b', '--batch',
            metavar = "BATCH_SIZE",
            help = "Batch size"
        )

        # define command using shared arg
        @use_shared_arg('batch')
        @command
        def my_command(batch: int = 5):
            assert batch == 25

        # check
        assert shared_arg_exists('batch')
        assert shared_arg_exists(batch)
        assert command_exists('my_command')
        assert command_exists(my_command)
        assert get_argument('my_command', 'batch') == {
            "kind":   "arg",
            "args":   tuple(['-b', '--batch']),
            "kwargs": {
                "metavar": "BATCH_SIZE",
                "help": "Batch size"
            },
            "name": "batch"
        }
        assert get_argument('my_command', batch) == {
            "kind":   "arg",
            "args":   tuple(['-b', '--batch']),
            "kwargs": {
                "metavar": "BATCH_SIZE",
                "help": "Batch size"
            },
            "name": "batch"
        }
        assert has_argument(my_command, batch)
        assert has_argument('my_command', 'batch')

    finally:
        # cleanup
        remove_command('my_command')
        remove_shared_arg(batch)

        assert not command_exists('my_command')
        assert not shared_arg_exists('batch')


# ------------------------------------------------------------------------------------------------------------------------
def test_cli_simple():
    """ Tests many functions of the simple cli system together. """
    try:
        # define a fruits arguments group
        @arg(
            '-b', '--banana',
            help    = 'Yummy yellow fruit',
            dest    = 'banana',
            metavar = 'BANANA',
            default = 'banana'
        )
        @arg(
            '-c', '--cherry',
            help    = 'Yummy red fruit',
            dest    = 'cherry',
            metavar = 'CHERRY',
            default = 'cherry'
        )
        @arg_group
        def fruits():
            pass

        # define a vegetables arguments group
        @arg(
            '-s', '--squash',
            help    = 'Yucky green vegetable',
            dest    = 'squash',
            metavar = 'SQUASH',
            default = 'squash'
        )
        @arg(
            '-t', '--tomato',
            help    = 'Yummy red vegetable',
            dest    = 'tomato',
            metavar = 'TOMATO',
            default = 'tomato'
        )
        @arg_group
        def vegetables():
            pass

        # define a produce arguments group
        @use_arg_group('fruits')
        @use_arg_group('vegetables')
        @arg_group
        def produce():
            pass

        # define a shared argument
        shared_arg(
            '-w', '--whatchamacallits',
            dest = "whatchamacallits"
        )

        # define a command using shared argument, argument group, and some directly defined arguments
        @use_shared_arg('whatchamacallits')
        @use_arg_group('produce')
        @arg(
            '-p', '--prefix',
            help    = 'Folder prefix',
            dest    = 'prefix',
            metavar = 'FOLDER',
            default = '/tmp'
        )
        @arg(
            '-f', '--force',
            help    = 'When present, ignore cached results and perform all actions again.',
            dest    = 'force',
            action  = 'store_true',
            default = False
        )
        @command
        def my_command(prefix: str, force: bool, **kwargs) -> None:
            print("I'm now running stuff!")

        assert arg_group_exists('fruits')
        assert has_argument('fruits', 'cherry')
        assert has_argument('fruits', 'banana')

        assert arg_group_exists('vegetables')
        assert has_argument('vegetables', 'squash')
        assert has_argument('vegetables', 'tomato')

        assert arg_group_exists('produce')
        assert has_argument(produce, 'cherry')
        assert has_argument('produce', 'banana')
        assert get_argument(produce, 'squash') is not None
        assert get_argument('produce', 'tomato') is not None

        assert shared_arg_exists('whatchamacallits')
        assert command_exists('my_command')

        assert has_argument('my_command', 'whatchamacallits')
        assert has_argument('my_command', 'cherry')
        assert has_argument('my_command', 'banana')
        assert has_argument('my_command', 'squash')
        assert has_argument('my_command', 'tomato')
        assert has_argument('my_command', 'force')
        assert has_argument('my_command', 'prefix')

        assert get_argument('my_command', 'whatchamacallits') == {
            "kind":   "arg",
            "args":   tuple(['-w', '--whatchamacallits']),
            "kwargs": {
                "dest": "whatchamacallits"
            },
            "name": "whatchamacallits"
        }

    finally:
        remove_shared_arg('whatchamacallits')
        remove_arg_group('produce')
        remove_arg_group('vegetables')
        remove_arg_group('fruits')
        remove_command('my_command')

        assert not command_exists('my_command')
        assert not arg_group_exists('produce')
        assert not arg_group_exists('vegetables')
        assert not arg_group_exists('fruits')
        assert not shared_arg_exists('whatchamacallits')



#!/bin/false
# -*- coding: utf-8 -*-

""" Tools to simplify spawning child processes and capturing, logging, or processing output. """


# =======================================================================================================================
# Main imports
# =======================================================================================================================
import os
import subprocess
import threading
import shlex
import time
import datetime
from enum import Enum
from typing import Any, Callable, Dict, Generator, IO, List, Optional, Tuple, Union
from queue import Queue, Empty
from pt.ptutils.time import dt, now
from pt.ptutils.typing import is_string


# ------------------------------------------------------------------------------------------------------------------------
# Logging config
# ------------------------------------------------------------------------------------------------------------------------
from pt.ptutils.logging import getLogger
logger = getLogger(__name__)


# =======================================================================================================================
# Globals
# =======================================================================================================================
""" Environment variables that are copied by default to child `Program` environments. """
__COPIED_ENV_ARGNAMES__ = [
    'LANG', 'LANGUAGE', 'LC_ALL',
    'DISPLAY', 'LS_COLORS',
    'HOST_TYPE',
    'USER', 'HOME', 'NAME', 'SHELL',
    'DOCKER_HOST', 'TERM', 'PATH'
]


# =======================================================================================================================
# Class: Program
# =======================================================================================================================
class Program(object):
    """ A child process. """

    # --------------------------------------------------------------------------------------------------------------------
    @staticmethod
    def run(
        args,
        cwd:             Optional[str]            = None,
        capture:         bool                     = True,
        display:         bool                     = True,
        raise_on_error:  bool                     = False,
        env:             Optional[Dict[str, str]] = None,
        timeout_seconds: Optional[float]          = None,
        silent:          bool                     = True
    ) -> Union[int, Tuple[int, List[str], List[str]] ]:
        """
        Convenience method to run a child process synchronously and await the result.

        Parameters
        ----------
        args : Union[str, List[str]]
            A list of program invocation arguments, or a string of such arguments separated by
            spaces. If providing a string, be aware that arguments containing spaces won't be
            handled appropriately and will be split into multiple arguments.
        cwd : Optional[str], optional
            When specified, change the current directory to the path provided, by default None
        capture : bool, optional
            When True, capture standard output and error streams of the child program. By default True
        display : bool, optional
            When True, log the standard output and error streams of the child program. By default True
        raise_on_error : bool, optional
            Raise an exception when the child process returns a return code other than 0, by default False
        env : Optional[Dict[str, Any]], optional
            A dictionary of environment arguments for the child program, by default None.
            When omitted, the child process inherits the parent environment as per
            `subprocess.Popen`.
            When present and `copy_shell_env` is True, the child environment will be the result
            of merging the supplied environment with a selected subset of the current process's
            environment. See `__COPIED_ENV_ARGNAMES__` to see what gets copied.
        timeout_seconds : float, optional
            When specified, kill the child process with SIGTERM if it hasn't completed in the
            specified number of seconds, by default None
        silent : bool, optional
            When true, disable logging about child process start/stop, except in the case of
            errors, by default True

        Returns
        -------
        Union[int, Tuple[int, List[str], List[str]] ]
            If `capture` is True, this will be a tuple of (return code, standard output line, standard error lines).
            If `capture` is False, this will be a just the return code.
        """
        if is_string(args):
            args = args.split(' ')
        p = Program(
            args,
            cwd             = cwd,
            capture_stdout  = capture,
            capture_stderr  = capture,
            display_stderr  = display,
            display_stdout  = display,
            silent          = silent,
            env             = env,
            timeout_seconds = timeout_seconds
        )
        p.start()
        rc = p.wait()
        if raise_on_error:
            if rc != 0:
                raise subprocess.CalledProcessError(rc, str(p), p.stdout, p.stderr)

        if capture:
            return (rc, p.stdout_lines, p.stderr_lines)
        else:
            return rc

    # --------------------------------------------------------------------------------------------------------------------
    class Channel(Enum):
        """ Identifies the output stream of a child processes. """

        """ The standard output stream. """
        STDOUT = 0

        """ The standard error stream. """
        STDERR = 1

    # --------------------------------------------------------------------------------------------------------------------
    def __str__(self) -> str:
        """
        String representation of the program invocation.

        Returns
        -------
        str
            [description]
        """
        return ' '.join([shlex.quote(arg) for arg in self.args])

    # --------------------------------------------------------------------------------------------------------------------
    def __repr__(self) -> str:
        """
        String representation of the program invocation.

        Returns
        -------
        str
            A terse string representation.
        """
        return "Program(%s)" % self

    # --------------------------------------------------------------------------------------------------------------------
    def __init__(
            self,
            args:            Union[str, List[str]],
            capture_stdout:  bool                                                        = True,
            capture_stderr:  bool                                                        = True,
            display_stdout:  bool                                                        = True,
            display_stderr:  bool                                                        = True,
            timeout_seconds: float                                                       = None,
            listener:        Optional[Callable[[Channel, str, datetime.datetime], None]] = None,
            env:             Optional[Dict[str, Any]]                                    = None,
            log_prefix:      Optional[str]                                               = None,
            silent:          bool                                                        = True,
            copy_shell_env:  bool                                                        = True,
            cwd:             Optional[str]                                               = None
    ):
        """
        A child process runner that captures and or logs outputs.

        Parameters
        ----------
        args : Union[str, List[str]]
            A list of program invocation arguments, or a string of such arguments separated by
            spaces. If providing a string, be aware that arguments containing spaces won't be
            handled appropriately and will be split into multiple arguments.
        capture_stdout : bool, optional
            When True, capture standard output stream of the child program. By default True
        capture_stderr : bool, optional
            When True, capture standard error stream of the child program. By default True
        display_stdout : bool, optional
            When True, log the standard output stream of the child program. By default True
        display_stderr : bool, optional
            When True, log the standard error stream of the child program. By default True
        timeout_seconds : float, optional
            When specified, kill the child process with SIGTERM if it hasn't completed in the
            specified number of seconds, by default None
        listener : Optional[Callable[[Channel, str, datetime.datetime], None]], optional
            A handler which, when specified, will be called with the channel, line content, and
            timestamp of each line of standard output or error streams of the child program,
            by default None
        env : Optional[Dict[str, Any]], optional
            A dictionary of environment arguments for the child program, by default None.
            When omitted, the child process inherits the parent environment as per
            `subprocess.Popen`.
            When present and `copy_shell_env` is True, the child environment will be the result
            of merging the supplied environment with a selected subset of the current process's
            environment. See `__COPIED_ENV_ARGNAMES__` to see what gets copied.
        log_prefix : Optional[str], optional
            An opptional line prefix for child log entries when logging child output, by
            default None
        silent : bool, optional
            When true, disable logging about child process start/stop, except in the case of
            errors, by default True
        copy_shell_env : bool, optional
            When specified and `env` is also provided, first copy a subset of the current
            environment and then update the child's environment with supplied entries in `env`,
            by default True. See `__COPIED_ENV_ARGNAMES__` to see what gets copied.
        cwd : Optional[str], optional
            When specified, change the current directory to the path provided, by default None
        """
        self.args             = [str(arg) for arg in args]
        self.capture_stdout   = capture_stdout
        self.capture_stderr   = capture_stderr
        self.display_stdout   = display_stdout
        self.display_stderr   = display_stderr
        self.timeout          = timeout_seconds
        self.listener         = listener
        self.env              = env
        self.log_prefix       = log_prefix
        self.silent           = silent
        self.cwd              = cwd

        self._proc            = None
        self._return_code     = None
        self._output          = []
        self._stopped         = threading.Event()
        self._stderr_thread   = None
        self._stdout_thread   = None
        self._handler_thread  = None
        self._watchdog_thread = None
        self._queue           = Queue()
        self._T_start         = None
        self.timed_out        = threading.Event()

        if (env is not None) and copy_shell_env:
            for k in __COPIED_ENV_ARGNAMES__:
                v = os.environ.get(k, None)
                if v is not None:
                    env[k] = v

    # --------------------------------------------------------------------------------------------------------------------
    def _watchdog(self) -> None:
        """
            Watchdog thread to kill the process if it runs longer than `self.timeout`.
        """
        if self.timeout is not None:
            while self._proc.poll() is None:
                t_now   = now()
                elapsed = dt(self._T_start, t_now)
                if elapsed > self.timeout:
                    logger.warning("Program failed to complete in %s seconds." % self.timeout)
                    self.timed_out.set()
                    self.terminate()
                    while self._proc.poll() is None:
                        time.sleep(0.01)
                else:
                    time.sleep(0.01)

    # --------------------------------------------------------------------------------------------------------------------
    def _pipe_reader(
        self,
        queue: Queue,
        pipe: IO[bytes],
        channel: Channel,
        capture: bool,
        display: bool
    ) -> None:
        """
        Generic thread to read froma pipe and insert into an inbox queue.
        This thread will automatically shutdown when the pipe is closed.

        Parameters
        ----------
        queue : Queue
            The queue into which to place the log entries.
        pipe : IO[bytes]
            The pipe to read from.
        channel : Channel
            The channel that the input pipe corresponds to.
        capture : bool
            Flag indicating that `channel` is configured to be captured.
        display : bool
            Flag indicating that `channel` is configured to be displayed.
        """
        try:
            with pipe:
                for line in iter(pipe.readline, b''):
                    queue.put(
                        (channel, capture, display, line, now())
                    )
        finally:
            queue.put(
                (channel, capture, display, None, now())
            )

    # --------------------------------------------------------------------------------------------------------------------
    def _handler(self, queue: Queue, stopped: threading.Event) -> None:
        """
        Output stream processing handler.
        Handles logging, capture, and dispatch to any custom handler.

        Parameters
        ----------
        queue : Queue
            The queue on which to receive stream output records
        stopped : threading.Event
            An event used to signal processing should be stopped.
        """
        prefix = (self.log_prefix + ' ') if self.log_prefix else ''
        while not stopped.isSet():
            try:
                (channel, capture, display, line, timestamp) = queue.get_nowait()
                if line is not None:
                    if isinstance(line, bytes):
                        line = line.decode('utf-8').rstrip()
                    if capture:
                        self._output.append((channel, line, timestamp))
                    if display:
                        logger.info( prefix + line )
                    if self.listener:
                        try:
                            self.listener(channel, line, timestamp)
                        except Exception:  # pylint: disable=broad-except
                            logger.exception("⚠️ Error during user-supplied message handler. (IGNORED)")
                queue.task_done()

            except Empty:
                time.sleep(0.05)

            except Exception:  # pylint: disable=broad-except
                logger.exception("🐞 Error processing program output")

    # --------------------------------------------------------------------------------------------------------------------
    def start(self) -> None:
        """
        Start the child program.
        """
        if not self.silent:
            logger.info("🟩 Starting program: %s" % self, data=self.env)

        self._T_start = now()

        # start the process
        self._proc = subprocess.Popen(
            self.args,
            stdout             = subprocess.PIPE,
            stderr             = subprocess.PIPE,
            bufsize            = 1,
            close_fds          = False,
            env                = self.env,
            universal_newlines = False,
            cwd                = self.cwd
            # ,
            # encoding           = 'utf-8'
        )

        # start the watchdog thread if required
        if self.timeout:
            self._watchdog_thread = threading.Thread( target=self._watchdog )
            self._watchdog_thread.start()

        # start the handler thread
        self._handler_thread = threading.Thread(
            target = self._handler,
            args   = [
                self._queue,
                self._stopped
            ]
        )
        self._handler_thread.start()

        # start the reader threads
        self._stdout_thread  = threading.Thread(
            target = self._pipe_reader,
            args   = [
                self._queue,
                self._proc.stdout,
                Program.Channel.STDOUT,
                self.capture_stdout,
                self.display_stdout
            ]
        )
        self._stdout_thread.start()

        self._stderr_thread  = threading.Thread(
            target = self._pipe_reader,
            args   = [
                self._queue,
                self._proc.stderr,
                Program.Channel.STDERR,
                self.capture_stderr,
                self.display_stderr
            ]
        )
        self._stderr_thread.start()

    # --------------------------------------------------------------------------------------------------------------------
    def wait(self) -> int:
        """
        Wait for the child program to complete and return it's return code.

        Returns
        -------
        int
            The child program's return code.
        """
        # wait for readers to see EOF
        self._stdout_thread.join()
        self._stderr_thread.join()
        self._queue.join()
        self._stopped.set()
        self._handler_thread.join()
        if self.timeout:
            self._watchdog_thread.join()
        self._return_code = self._proc.wait()
        if not self.silent:
            logger.info("🟥 Program completed: %s --> %s" % (self, self._return_code))
        return self._return_code

    # --------------------------------------------------------------------------------------------------------------------
    def send_signal(self, sig: int) -> None:
        """
        Send a signal to the child process.

        Parameters
        ----------
        sig : int
            The signal to send.
        """
        if not self.silent:
            logger.info("🔥 Terminating program: %s with signal ⚡%s⚡" % (self, sig))
        self._proc.send_signal(sig)

    # --------------------------------------------------------------------------------------------------------------------
    def terminate(self) -> None:
        """
        Terminate the child process by sending SIGTERM
        """
        if not self.silent:
            logger.info("🔥 Terminating program: %s" % self)
        self._proc.terminate()

    # --------------------------------------------------------------------------------------------------------------------
    @property
    def return_code(self) -> Optional[int]:
        """
        The program's return code, or return None if the program is still running.

        Returns
        -------
        Optional[int]
            The return code.
        """
        if self._return_code is None:
            self._return_code = self._proc.poll()
        return self._return_code

    # --------------------------------------------------------------------------------------------------------------------
    @property
    def is_finished(self) -> bool:
        """
        Test if the child program has completed.

        Returns
        -------
        bool
            True if the child program is still running, False otherwise.
        """
        return self.return_code is not None

    # --------------------------------------------------------------------------------------------------------------------
    @property
    def records(self) -> Generator[Tuple[Channel, str, datetime.datetime], None, None]:
        """
        Captured output records from either of the child program's streams.

        Yields
        -------
        Tuple[Channel, str, datetime.datetime]
            A record for each line captured from the child process as a tuple (channel, line, timestamp).
        """
        for x in self._output:
            yield x

    # --------------------------------------------------------------------------------------------------------------------
    @property
    def lines(self) -> Generator[str, None, None]:
        """
        Captured output lines from either of the child program's streams.
        Generator yielding individual output lines from either channel of the child.

        Yields
        -------
        str
            A line for each captured line on either stream of the child program.
        """
        for (_, line, __) in self._output:
            yield line

    # --------------------------------------------------------------------------------------------------------------------
    @property
    def stdout_lines(self) -> Generator[str, None, None]:
        """
        Captured output lines from the child program's standard output stream.

        Yields
        -------
        str
            A line for each captured line on the child program's standard output stream.
        """
        for (channel, line, _) in self._output:
            if channel == Program.Channel.STDOUT:
                yield line

    # --------------------------------------------------------------------------------------------------------------------
    @property
    def stderr_lines(self) -> Generator[str, None, None]:
        """
        Captured output lines from the child program's standard error stream.

        Yields
        -------
        str
            A line for each captured line on the child program's standard error stream.
        """
        for (channel, line, _) in self._output:
            if channel == Program.Channel.STDERR:
                yield line

    # --------------------------------------------------------------------------------------------------------------------
    @property
    def output(self) -> str:
        """
        The combined output of the child process as a linux newline-terminated block of text.

        Returns
        -------
        str
            The combined output.
        """
        return '\n'.join(self.lines)

    # --------------------------------------------------------------------------------------------------------------------
    @property
    def stdout(self) -> str:
        """
        The output of the child process's standard output stream as a linux newline-terminated block of text.

        Returns
        -------
        str
            The output.
        """
        return '\n'.join(self.stdout_lines)

    # --------------------------------------------------------------------------------------------------------------------
    @property
    def stderr(self) -> str:
        """
        The output of the child process's standard error stream as a linux newline-terminated block of text.

        Returns
        -------
        str
            The output.
        """
        return '\n'.join(self.stderr_lines)



#!/bin/false
# -*- coding: utf-8 -*-


""" Legacy python2 file io utility functions. This module is dated and will eventually be replaced by object-oriented
versions of these utilities. """

# =======================================================================================================================
# Main imports
# =======================================================================================================================
import os
import tempfile
import errno
from typing import Any, Union, Callable, List
from pt.ptutils.encoding import encode_json, decode_json, encode_yaml, decode_yaml, require_yaml
from pt.ptutils.undefined import UNDEFINED, Undefined, is_undefined

# =======================================================================================================================
# Import checks
# =======================================================================================================================
try:
    import tarfile
    HAVE_TAR = True
except ImportError:
    HAVE_TAR = False

# -----------------------------------------------------------------------------------------------------------------------
# Globals
# -----------------------------------------------------------------------------------------------------------------------
__FILETYPES__ = {
    'archive-zip': ['zip'],
    'archive-tar': ['tar'],
    'archive-bzip2': ['bz2'],
    'archive-gzip':  ['gz'],
    'yaml': ['yml', 'yaml'],
    'json': ['jsn', 'json']
}


# -----------------------------------------------------------------------------------------------------------------------
# Utility functions
# -----------------------------------------------------------------------------------------------------------------------
def get_path_part(filename: str, part: str) -> str:
    """
    Get the specified part of a filename.

    Parameters
    ----------
    filename : str
        The path or filename as a string.
    part : str
        The desired part. Valid part names are: 'fullpath', 'abspath', 'name', 'extension', 'basename'

    Returns
    -------
    str
        The indicated part.

    Raises
    ------
    ValueError
        If the specified part type is not one of the known types.
    """
    if part == 'fullpath':
        return filename
    elif part == 'abspath':
        return os.path.abspath(filename)
    elif part == 'name':
        return os.path.split(filename.rstrip(os.sep))[1]
    elif part == 'extension':
        (_, e) = os.path.splitext(
            os.path.split(
                filename.rstrip(os.sep)
            )[1]
        )
        return e.lstrip('.')
    elif part == 'basename':
        (b, _) = os.path.splitext(
            os.path.split(
                filename.rstrip(os.sep)
            )[1]
        )
        return b
    else:
        raise ValueError(f"Unknown filename part: '{part}'")


# ------------------------------------------------------------------------------------------------------------------------
def get_hash_tmp_dir(relpath: str) -> str:
    """
    Get a temporary directory based on a hashing of another directory name.

    Parameters
    ----------
    relpath : str
        The other directory name or path to hash.

    Returns
    -------
    str
        A temporary directory name as a string.
    """
    return os.path.join( tempfile.gettempdir(), hash(relpath) )


# ------------------------------------------------------------------------------------------------------------------------
def get_filetype(filename: str) -> str:
    """
    Get the registered filetype based on the file's extension.

    Parameters
    ----------
    filename : str
        The path or filename.

    Returns
    -------
    str
        The registered typename. See `__FILETYPES__` for more a list of known filetypes.
    """
    global __FILETYPES__
    extension = get_path_part(filename, 'extension').lower()
    for (filetype, extensions) in __FILETYPES__.items():
        if extension in extensions:
            return filetype
    return extension.lstrip('.')


# ------------------------------------------------------------------------------------------------------------------------
def makedirs(path: str) -> None:
    """
    Create a directory, recursively creating any parents as needed.

    Parameters
    ----------
    path : str
        The path to the folder to create. When relative, it is assumed to be relative to the current
        working dir (`os.curdir`).

    Raises
    ------
    OSError
        As per `os.makedirs`, except in the case of `errno.EEXIST`, in which case no error is raised.

    """
    try:
        os.makedirs(path)
    except OSError as exc:  # Python >2.5
        if not (exc.errno == errno.EEXIST and os.path.isdir(path)):
            raise


# ------------------------------------------------------------------------------------------------------------------------
def read(filename: str) -> str:
    """
    Read the entire content of a text file.

    Parameters
    ----------
    filename : str
        The name of the file.

    Returns
    -------
    str
        The content of the file as a string.

    Raises
    ------
    Exception
        As per `open` or `io.TextIOBase.read`.

    """
    with open(filename, 'r') as filp:
        return filp.read()


# ------------------------------------------------------------------------------------------------------------------------
def read_lines(filename: str) -> List[str]:
    """
    Read all of the lines of a text file.

    Parameters
    ----------
    filename : str
        The path to the file.

    Returns
    -------
    List[str]
        A list of the lines in the text file.

    Raises
    ------
    Exception
        As per `open` or `io.TextIOBase.readlines`.
    """
    with open(filename, 'r') as filp:
        return [ x.rstrip('\r\n') for x in filp.readlines() ]


# ------------------------------------------------------------------------------------------------------------------------
def read_all(filename: str) -> str:
    """
    Read the entire content of a text file.

    Parameters
    ----------
    filename : str
        The name of the file.

    Returns
    -------
    str
        The content of the file as a string.

    Raises
    ------
    Exception
        As per `open` or `io.TextIOBase.read`.


    Notes
    -----
    This is an alias of `read`.
    """
    return read(filename)


# ------------------------------------------------------------------------------------------------------------------------
def write_all(filename: str, content: str) -> None:
    """
    Write the content of a string to a text file.

    Parameters
    ----------
    filename : str
        The name of the file.
    content: str
        The text content of the file to be written.

    Raises
    ------
    Exception
        As per `open` or `io.TextIOBase.write`.

    """
    with open(filename, 'w') as filp:
        filp.write(content)


# ------------------------------------------------------------------------------------------------------------------------
def require_tar() -> None:
    """
    Test if TarFile support is available or raise an Exception.

    Raises
    ------
    Exception
        When `import tarfile` raises `ImportError`
    """
    if not HAVE_TAR:
        raise Exception("Tarfile support is not available on this platform.")


# ------------------------------------------------------------------------------------------------------------------------
def extract_archive(archive: str, folder: str) -> None:
    """
    Extract an archive file to the specified folder.

    Parameters
    ----------
    archive : str
        Path to the archive file to extract.
    folder : str
        Path to the folder in which to extract the archive's contents.

    Raises
    ------
    NotImplementedError
        When the archive pointed to by `archive` is of an unsupported filetype.
    Exception
        when tarfile support is unavailable on this platform.
    Exception
        As per `tarfile.open` or `tarfile.TarFile.extractall`
    """
    filetype = get_filetype(archive)
    if filetype == 'archive-zip':
        raise NotImplementedError('No support for .zip archives')
    elif filetype in ['archive-tar', 'archive-gzip', 'archive-bzip2']:
        require_tar()
        tar = tarfile.open(archive)
        tar.extractall(folder)
        tar.close()
    else:
        raise NotImplementedError(f"No extraction support for files of type '{filetype}'.")


# ------------------------------------------------------------------------------------------------------------------------
def load_txt(filename: str) -> str:
    """
    Read the entire content of a text file.

    Parameters
    ----------
    filename : str
        The name of the file.

    Returns
    -------
    str
        The content of the file as a string.

    Raises
    ------
    Exception
        As per `open` or `io.TextIOBase.read`.

    Notes
    -----
    This is an alias of `read_all`.

    """
    return read_all(filename)


# ------------------------------------------------------------------------------------------------------------------------
def save_txt(filename: str, obj: Any) -> None:
    """
    Write the content of a string to a text file.

    Parameters
    ----------
    filename : str
        The name of the file.
    obj: Any
        A string or object to write to the file. If not a string, `obj` will be converted
        to string by `str`.

    Raises
    ------
    Exception
        As per `open` or `io.TextIOBase.write`.
    """
    write_all(filename, str(obj))


# ------------------------------------------------------------------------------------------------------------------------
def load_json(filename: str) -> Any:
    """
    Load and deserialize a JSON-formatted text file.

    Parameters
    ----------
    filename : str
        The path to the file.

    Returns
    -------
    Any
        The python-native object represented by the content of the JSON file. See `decode_json`
        for more information on how `obj` is decoded.
    """
    return decode_json(read_all(filename))


# ------------------------------------------------------------------------------------------------------------------------
def save_json(filename: str, obj: Any) -> None:
    """
    Serialize and save a python object to a JSON-formatted text file.

    Parameters
    ----------
    filename : str
        The path to the file.
    obj : Any
        The object to serialize. Must consist of JSON serializable types. See `encode_json` for more
        information on how `obj` is encoded.
    """
    write_all(filename, encode_json(obj))


# ------------------------------------------------------------------------------------------------------------------------
def load_yaml(filename: str) -> Any:
    """
    Load and deserialize a YAML-formatted textfile.

    Parameters
    ----------
    filename : str
        The path to the file.

    Returns
    -------
    Any
        The python-native object represented by the content of the YAML file. See `decode_yaml`
        for more information on how `obj` is decoded.
    """
    with open(filename, 'r') as filp:
        return decode_yaml(filp)


# ------------------------------------------------------------------------------------------------------------------------
def save_yaml(filename: str, obj: Any) -> None:
    """
    Serialize and save a python object to a YAML-formatted text file.

    Parameters
    ----------
    filename : str
        The path to the file.
    obj : Any
        The object to serialize. Must consist of JSON serializable types. See `encode_yaml` for more
        information on how `obj` is encoded.
    """
    yaml = require_yaml()
    with open(filename, 'w') as filp:
        yaml.safe_dump(obj, filp, explicit_start=True, default_flow_style=False)


# ------------------------------------------------------------------------------------------------------------------------
def get_encoder(filename: str) -> Callable[[Any], str]:
    """
    Get the most appropriate encoder for serializing data to a format consistent with
    the file's extension.

    Parameters
    ----------
    filename : str
        The filename to examine.

    Returns
    -------
    Callable[[Any],str]
        A function which encodes an object to the text format most appropriate to the file's extension.

    Raises
    ------
    Exception
        When the filename's extension is not recognized. (see `__FILETYPES__` for more information.)
    """

    filetype = get_filetype(filename)
    if filetype == 'yaml':
        return encode_yaml

    elif filetype == 'json':
        return encode_json

    else:
        raise Exception(f'No known encoder for files with extension: "{filetype}"')


# ------------------------------------------------------------------------------------------------------------------------
def get_decoder(filename: str) -> Callable[ [str], Any ]:
    """
    Get the most appropriate decoder for serializing data to a format consistent with
    the file's extension.

    Parameters
    ----------
    filename : str
        The filename to examine.

    Returns
    -------
    Callable[ [str], Any ]
        A function which decodes an object from the text format most appropriate to the file's extension.

    Raises
    ------
    Exception
        When the filename's extension is not recognized. (see `__FILETYPES__` for more information.)
    """
    filetype = get_filetype(filename)
    if filetype == 'yaml':
        return decode_yaml

    elif filetype == 'json':
        return decode_json

    else:
        raise Exception(f'No known decoder for files with extension: "{filetype}"')


# ------------------------------------------------------------------------------------------------------------------------
def get_loader(
    filename: str,
    default:  Union[Undefined, Callable[[str], Any]] = UNDEFINED
) -> Callable[[str], Any]:
    """
    Get the most appropriate file loader for reading and deserializing data from a format consistent with
    the file's extension.

    Parameters
    ----------
    filename : str
        The filename to examine.
    default : Union[Undefined, Callable[[str], Any]], optional
        A loader function to return if no other loader could be found, by default UNDEFINED

    Returns
    -------
    Callable[[str],Any]
        A function which takes a filename and returns the deserialized content of that file.

    Raises
    ------
    Exception
        When no known loader could be found and `default` is `UNDEFINED`.
    """
    filetype = get_filetype(filename)
    if filetype == 'yaml':
        return load_yaml

    elif filetype == 'json':
        return load_json

    else:
        if is_undefined(default):
            raise Exception(f'No known load function for files with extension: "{filetype}"')
        else:
            return default


# ------------------------------------------------------------------------------------------------------------------------
def get_saver(
    filename: str,
    default: Union[Undefined, Callable[[str, Any], None]] = UNDEFINED
) -> Callable[[str, Any], None]:
    """
    Get the most appropriate file saver for serializing and writing data in a format consistent with
    the file's extension.

    Parameters
    ----------
    filename : str
        The filename to examine.
    default : Union[Undefined, Callable[[str, Any],None]], optional
        A saver function to return if no other saver could be found, by default UNDEFINED

    Returns
    -------
    Callable[[str, Any],None]
        A function which takes a filename and an object and writes the serialized object to that file.

    Raises
    ------
    Exception
        When no known saver could be found and `default` is `UNDEFINED`.
    """
    filetype = get_filetype(filename)
    if filetype == 'yaml':
        return save_yaml

    elif filetype == 'json':
        return save_json

    else:
        if is_undefined(default):
            raise Exception(f'No known save function for files with extension: "{filetype}"')
        else:
            return default


# ------------------------------------------------------------------------------------------------------------------------
def load(filename: str) -> Any:
    """
    Load the data in a file using an appropriate loader function based on the file's filename.

    Parameters
    ----------
    filename : str
        The filename to read and deserialize.

    Returns
    -------
    Any
        The deserialized object. Deserialization of this object must be supported by the
        relevant loader function. (See `get_loader` for more information).
    """
    return get_loader(filename)(filename)


# ------------------------------------------------------------------------------------------------------------------------
def save(filename: str, obj: Any) -> None:
    """
    Save a python object to a data file in a format determined by the file's filename.

    Parameters
    ----------
    filename : str
        The filename into which serialized version of `obj` will be written.
    obj : [type]
        The object to serialize. Serialization of this object must be supported by the
        relevant saver function. (See `get_saver` for more information).
    """
    get_saver(filename)(filename, obj)
